/*
 *
 *    ******************************************************************************
 *    *
 *    * This program and the accompanying materials are made available under the
 *    * terms of the Apache License, Version 2.0 which is available at
 *    * https://www.apache.org/licenses/LICENSE-2.0.
 *    *
 *    *  See the NOTICE file distributed with this work for additional
 *    *  information regarding copyright ownership.
 *    * Unless required by applicable law or agreed to in writing, software
 *    * distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
 *    * WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
 *    * License for the specific language governing permissions and limitations
 *    * under the License.
 *    *
 *    * SPDX-License-Identifier: Apache-2.0
 *    *****************************************************************************
 *
 */

package net.brutex.gan;



import java.awt.*;
import java.awt.image.BufferedImage;
import java.io.File;
import java.io.IOException;
import java.util.*;
import java.util.List;
import javax.imageio.ImageIO;
import javax.swing.*;
import lombok.extern.slf4j.Slf4j;
import org.datavec.api.split.FileSplit;
import org.datavec.image.loader.NativeImageLoader;
import org.datavec.image.recordreader.ImageRecordReader;
import org.datavec.image.transform.*;
import org.deeplearning4j.datasets.datavec.RecordReaderDataSetIterator;
import org.deeplearning4j.datasets.iterator.impl.MnistDataSetIterator;
import org.deeplearning4j.nn.conf.GradientNormalization;
import org.deeplearning4j.nn.conf.NeuralNetConfiguration;
import org.deeplearning4j.nn.conf.layers.*;
import org.deeplearning4j.nn.conf.layers.misc.FrozenLayerWithBackprop;
import org.deeplearning4j.nn.multilayer.MultiLayerNetwork;
import org.deeplearning4j.optimize.listeners.PerformanceListener;
import org.junit.jupiter.api.Tag;
import org.junit.jupiter.api.Test;
import org.nd4j.linalg.api.ndarray.INDArray;
import org.nd4j.linalg.dataset.DataSet;
import org.nd4j.linalg.dataset.api.iterator.DataSetIterator;
import org.nd4j.linalg.factory.Nd4j;

@Slf4j
public class App2 {

    final int INPUT = CHANNELS*DIMENSIONS*DIMENSIONS;
    static final float COLORSPACE = 255f;
    static final int DIMENSIONS = 28;
    static final int CHANNELS = 1;
    final int ARRAY_SIZE_PER_SAMPLE = DIMENSIONS*DIMENSIONS*CHANNELS;
    final int OUTPUT_PER_PANEL = 10;

    final boolean BIAS = true;

    static final int BATCHSIZE=128;

    private JFrame frame2, frame;
    static final String OUTPUT_DIR = "d:/out/";

    final static INDArray label_real = Nd4j.ones(BATCHSIZE, 1);
    final static INDArray label_fake = Nd4j.zeros(BATCHSIZE, 1);

    @Test @Tag("long-running")
    void runTest() throws IOException {
        Nd4j.getMemoryManager().setAutoGcWindow(15 * 1000);

        MnistDataSetIterator mnistIter = new MnistDataSetIterator(20, 200);
        FileSplit fileSplit = new FileSplit(new File("c:/users/brian/downloads/humans2"), NativeImageLoader.getALLOWED_FORMATS());
        ImageTransform transform = new ColorConversionTransform(new Random(42), 7 );
        ImageTransform transform2 = new ShowImageTransform("Tester", 30);
        ImageTransform transform3 = new ResizeImageTransform(DIMENSIONS, DIMENSIONS);

        ImageTransform tr = new PipelineImageTransform.Builder()
               .addImageTransform(transform) //convert to GREY SCALE
                .addImageTransform(transform3)
                //.addImageTransform(transform2)
                .build();

        ImageRecordReader imageRecordReader = new ImageRecordReader(DIMENSIONS, DIMENSIONS, CHANNELS);
        imageRecordReader.initialize(fileSplit, tr);
        DataSetIterator trainData = new RecordReaderDataSetIterator(imageRecordReader, BATCHSIZE );
        trainData = new MnistDataSetIterator(BATCHSIZE, true, 42);

        MultiLayerNetwork dis = new MultiLayerNetwork(App2Config.discriminator());
        MultiLayerNetwork gen = new MultiLayerNetwork(App2Config.generator());

        LayerConfiguration[] disLayers = App2Config.discriminator().getFlattenedLayerConfigurations().stream()
                .map((layer) -> {
                    if (layer instanceof DenseLayer || layer instanceof OutputLayer) {
                        return FrozenLayerWithBackprop.builder(layer).name("frozen-for-"+layer.getName()).build();
                    } else {
                        return layer;
                    }
                }).toArray(LayerConfiguration[]::new);

    NeuralNetConfiguration netConfiguration =
        NeuralNetConfiguration.builder()
            .name("GAN")
            .gradientNormalization(GradientNormalization.RenormalizeL2PerLayer)
                .gradientNormalizationThreshold(100)
                .updater(App2Config.UPDATER)
            .innerConfigurations(new ArrayList<>(List.of(App2Config.generator())))
            .layersFromList(new ArrayList<>(Arrays.asList(disLayers)))
            // .inputType(InputType.convolutional(DIMENSIONS, DIMENSIONS, CHANNELS))
            // .inputPreProcessor(4, new CnnToFeedForwardPreProcessor())
            //.inputPreProcessor(0, new CnnToFeedForwardPreProcessor())
           // .inputPreProcessor(2, new FeedForwardToCnnPreProcessor(DIMENSIONS, DIMENSIONS, CHANNELS))
                //.inputPreProcessor(2, new CnnToFeedForwardPreProcessor())
                //.dataType(DataType.FLOAT)
            .build();

        MultiLayerNetwork gan = new MultiLayerNetwork(netConfiguration );

        dis.init(); log.debug("Discriminator network: {}", dis);
        gen.init(); log.debug("Generator network: {}", gen);
        gan.init(); log.debug("GAN network: {}", gan);


        log.info("Generator Summary:\n{}", gen.summary());
        log.info("GAN Summary:\n{}", gan.summary());
        dis.addTrainingListeners(new PerformanceListener(10, true, "DIS"));
        gen.addTrainingListeners(new PerformanceListener(10, true, "GEN"));
        gan.addTrainingListeners(new PerformanceListener(10, true, "GAN"));

        int j = 0;
        for (int i = 0; i < 51; i++) { //epoch
            while (trainData.hasNext()) {
                j++;
                DataSet next = trainData.next();
                // generate data
                INDArray real = next.getFeatures(); //.muli(2).subi(1);;//.div(255f);

                //start next round if there are not enough images left to have a full batchsize dataset
                if(real.length() < ARRAY_SIZE_PER_SAMPLE*BATCHSIZE) {
                    log.warn("Your total number of input images is not a multiple of {}, "
                            + "thus skipping {} images to make it fit", BATCHSIZE, real.length()/ARRAY_SIZE_PER_SAMPLE);
                    break;
                }

                //if(i%20 == 0) {

               // frame2 = visualize(new INDArray[]{real}, BATCHSIZE,
               //         frame2 == null ? new JFrame() : frame2, true); //real has batchsize number of images
                //}
                //real.divi(255f);

//        int batchSize = (int) real.shape()[0];

                //INDArray fakeIn = Nd4j.rand(BATCHSIZE, CHANNELS, DIMENSIONS, DIMENSIONS);
                //INDArray fakeIn = Nd4j.rand(new int[]{batchSize, X_DIM*Y_DIM}); //hack for MNIST only, use above otherwise
                INDArray fakeIn = Nd4j.rand(BATCHSIZE, App2Config.INPUT);

                INDArray fake = gan.activateSelectedLayers(0, gen.getLayers().length - 1, fakeIn);
                // when generator has TANH as activation - value range is -1 to 1
                // when generator has SIGMOID, then range is 0 to 1
                fake.addi(1f).divi(2f);

                DataSet realSet = new DataSet(real, label_real);
                DataSet fakeSet = new DataSet(fake, label_fake);

                DataSet data = DataSet.merge(Arrays.asList(realSet, fakeSet));

                dis.fit(data);
                dis.fit(data);
                // Update the discriminator in the GAN network
                updateGan(gen, dis, gan);

                gan.fit(new DataSet(Nd4j.rand(BATCHSIZE, App2Config.INPUT), label_fake));

                //Visualize and reporting
                if (j % 10 == 1) {
                    System.out.println("Epoch " + i + " Iteration " + j + " Visualizing...");
                    INDArray[] samples = BATCHSIZE > OUTPUT_PER_PANEL ? new INDArray[OUTPUT_PER_PANEL] : new INDArray[BATCHSIZE];


                    for (int k = 0; k < samples.length; k++) {
                        DataSet fakeSet2 = new DataSet(fakeIn, label_fake);
                        INDArray input = fakeSet2.get(k).getFeatures();

                        //input = input.reshape(1,CHANNELS, DIMENSIONS, DIMENSIONS); //batch size will be 1 here for images
                        input = input.reshape(1, App2Config.INPUT);

                        //samples[k] = gen.output(input, false);
                        samples[k] = gen.activateSelectedLayers(0, gen.getLayers().length - 1, input);
                        samples[k] = samples[k].reshape(1, CHANNELS, DIMENSIONS, DIMENSIONS);
                        //samples[k] =
                        //samples[k].muli(255f);

                    }
                    frame = visualize(samples, 1, frame == null ? new JFrame() : frame, false); //each samples only has 1 image, thus batchElements=1
                }
            }

            if (trainData.resetSupported()) {
                trainData.reset();
            } else {
                log.error("Trainingdata {} does not support reset.", trainData.toString());
            }
            // Copy the GANs generator to gen.
            updateGen(gen, gan);
            log.info("Updated GAN's generator from gen.");
            gen.save(new File("mnist-mlp-generator.dlj"));
        }
    }











    private static JFrame visualize(INDArray[] samples, int batchElements, JFrame frame, boolean isOrig) {
        if (isOrig) {
            frame.setTitle("Viz Original");
        } else {
            frame.setTitle("Generated");
        }

        frame.setDefaultCloseOperation(WindowConstants.DISPOSE_ON_CLOSE);
        frame.setLayout(new BorderLayout());

        JPanel panelx = new JPanel();

        panelx.setLayout(new GridLayout(4, 4, 8, 8));
        for (INDArray sample : samples) {
            for(int i = 0; i<batchElements; i++) {
                panelx.add(getImage(sample, i, isOrig));
            }
        }
        frame.add(panelx, BorderLayout.CENTER);
        frame.setVisible(true);

        frame.revalidate();
        frame.setMinimumSize(new Dimension(300, 20));
        frame.pack();
        return frame;
    }


    private static JLabel getImage(INDArray tensor, int batchElement, boolean isOrig) {
        final BufferedImage bi;
        if(CHANNELS >1) {
            bi = new BufferedImage(DIMENSIONS, DIMENSIONS, BufferedImage.TYPE_INT_RGB); //need to change here based on channels
        } else {
            bi = new BufferedImage(DIMENSIONS, DIMENSIONS, BufferedImage.TYPE_BYTE_GRAY); //need to change here based on channels
        }
        final int imageSize = DIMENSIONS * DIMENSIONS;
        final int offset = batchElement * imageSize;
        int pxl = offset * CHANNELS; //where to start in the INDArray

        //Image in NCHW - channels first format
        for (int c = 0; c < CHANNELS; c++) { //step through the num channels for each pixel
            for (int y = 0; y < DIMENSIONS; y++) { // step through the columns x
                for (int x = 0; x < DIMENSIONS; x++) { //step through the rows y
                    float f_pxl = tensor.getFloat(pxl) * COLORSPACE;
                    if(isOrig) log.trace("'{}.' Image (x,y,c): ({}, {}, {}) with INDArray with index {} and value '{}'", batchElement, x, y, c, pxl, f_pxl);
                    bi.getRaster().setSample(x, y, c, f_pxl);
                    pxl++; //next item in INDArray
                }
            }
        }
        ImageIcon orig = new ImageIcon(bi);
        Image imageScaled = orig.getImage().getScaledInstance((4 * DIMENSIONS), (4 * DIMENSIONS), Image.SCALE_DEFAULT);
        ImageIcon scaled = new ImageIcon(imageScaled);
        if(! isOrig)  saveImage(imageScaled, batchElement, isOrig);
        return new JLabel(scaled);

    }

    private static void saveImage(Image image, int batchElement, boolean isOrig) {
        String outputDirectory = OUTPUT_DIR; // Set the output directory where the images will be saved

        try {
            // Save the images to disk
            saveImage(image, outputDirectory, UUID.randomUUID().toString()+".png");

            log.debug("Images saved successfully.");
        } catch (IOException e) {
            log.error("Error saving the images: {}", e.getMessage());
        }
    }
    private static void saveImage(Image image, String outputDirectory, String fileName) throws IOException {
        File directory = new File(outputDirectory);
        if (!directory.exists()) {
            directory.mkdir();
        }

        File outputFile = new File(directory, fileName);
        ImageIO.write(imageToBufferedImage(image), "png", outputFile);
    }

    public static BufferedImage imageToBufferedImage(Image image) {
        if (image instanceof BufferedImage) {
            return (BufferedImage) image;
        }

    // Create a buffered image with the same dimensions and transparency as the original image
        BufferedImage bufferedImage;
    if (CHANNELS > 1) {
      bufferedImage =
          new BufferedImage(
              image.getWidth(null), image.getHeight(null), BufferedImage.TYPE_INT_ARGB);
        } else {
        bufferedImage =
                new BufferedImage(
                        image.getWidth(null), image.getHeight(null), BufferedImage.TYPE_BYTE_GRAY);
        }

        // Draw the original image onto the buffered image
        Graphics2D g2d = bufferedImage.createGraphics();
        g2d.drawImage(image, 0, 0, null);
        g2d.dispose();

        return bufferedImage;
    }
    private static void updateGen(MultiLayerNetwork gen, MultiLayerNetwork gan) {
        for (int i = 0; i < gen.getLayers().length; i++) {
            gen.getLayer(i).setParams(gan.getLayer(i).getParams());
        }
    }

    private static void updateGan(MultiLayerNetwork gen, MultiLayerNetwork dis, MultiLayerNetwork gan) {
        int genLayerCount = gen.getLayers().length;
        for (int i = genLayerCount; i < gan.getLayers().length; i++) {
            gan.getLayer(i).setParams(dis.getLayer(i - genLayerCount).getParams());
        }
    }

}
